﻿using System;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Text;
using System.Xml;

namespace NetworkChallengedStreams
{
    class Client
    {
        public static void Run(int serverTestScenario, int clientTestScenario)
        {
            try
            {
                Console.WriteLine("Connecting to 127.0.0.1:{0}...", Constants.Port);

                var endPoint = new IPEndPoint(IPAddress.Loopback, Constants.Port);

                using (var socket = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp))
                {
                    socket.NoDelay = true;
                    socket.Connect(endPoint);

                    Console.WriteLine("Executing client scenario {0}, server scenario {1}", clientTestScenario, serverTestScenario);

                    using (var stream = new NetworkStream(socket))
                    {
                        stream.WriteByte((byte)serverTestScenario);
                        stream.Flush();

                        ProcessResponse(clientTestScenario, new LoggingStream(stream)); // replace with just "stream" to turn off binary logging
                        stream.WriteByte(1);
                        stream.Flush();
                        socket.Shutdown(SocketShutdown.Send);
                    }
                }
            }
            catch (Exception ex)
            {
                Console.WriteLine("Client error: " + ex);
            }

        }

        private static void ProcessResponse(int clientTestScenario, Stream stream)
        {
            switch (clientTestScenario)
            {
                case 1:
                    ReadBinary(stream, 512);
                    break;

                case 2:
                    ReadText(stream, 512);
                    break;

                case 3:
                    ReadText(stream, 1024);
                    break;

                case 4:
                    ReadXmlDirectly(stream);
                    break;

                case 5:
                    ReadXmlFromTextReader(stream);
                    break;

                case 6:
                    ReadXmlSubtree(stream);
                    break;

                case 7:
                    ReadXmlSubtreeFromTextReader(stream);
                    break;

                default:
                    throw new ApplicationException("Client test scenario not supported: " + clientTestScenario);
            }
        }

        private static void ReadBinary(Stream stream, int bufferSize)
        {
            Console.WriteLine("Reading input as binary. Buffer size is " + bufferSize);

            var buffer = new byte[bufferSize];
            while (true)
            {
                var bytes = stream.Read(buffer, 0, buffer.Length);
                if (bytes == 0) return;
                var s = Encoding.ASCII.GetString(buffer, 0, bytes);
                Console.Write(s);
                if (s.EndsWith("</message>")) return;
            }
        }

        private static void ReadText(Stream stream, int bufferSize)
        {
            Console.WriteLine("Reading input as text. Buffer size is " + bufferSize);

            var buffer = new char[bufferSize];

            using (var reader = new LoggingTextReader(new StreamReader(stream)))
            while (true)
            {
                var chars = reader.Read(buffer, 0, buffer.Length);
                if (chars == 0) return;
                var s = new string(buffer, 0, chars);
                Console.Write(s);
                if (s.EndsWith("</message>")) return;
            }
        }

        private static void ReadXmlDirectly(Stream stream)
        {
            Console.WriteLine("Reading input as XML");

            var doc = new XmlDocument();
            doc.Load(stream);
            Console.WriteLine("Received XML document: {0}", doc.OuterXml);
        }

        private static void ReadXmlFromTextReader(Stream stream)
        {
            Console.WriteLine("Reading input as XML via text reader");

            using (var textReader = new LoggingTextReader(new StreamReader(stream, new UTF8Encoding(false))))
            {
                var doc = new XmlDocument();
                doc.Load(textReader);
                Console.WriteLine("Received XML document: {0}", doc.OuterXml);
            }
        }

        private static void ReadXmlSubtree(Stream stream)
        {
            Console.WriteLine("Reading input as XML subtree");

            using (var xmlReader = new XmlTextReader(stream))
            {
                while (xmlReader.Read())
                {
                    if (xmlReader.NodeType != XmlNodeType.Element) continue;
                    using (var subTreeReader = xmlReader.ReadSubtree())
                    {
                        var doc = new XmlDocument();
                        doc.Load(subTreeReader);
                        Console.WriteLine("Received XML document: {0}", doc.OuterXml);
                        return;
                    }
                }
            }
        }

        private static void ReadXmlSubtreeFromTextReader(Stream stream)
        {
            Console.WriteLine("Reading input as XML subtree");

            using (var textReader = new LoggingTextReader(new StreamReader(stream, new UTF8Encoding(false), false, 4096)))
            using (var xmlReader = new XmlTextReader(textReader))
            {
                while (xmlReader.Read())
                {
                    if (xmlReader.NodeType != XmlNodeType.Element) continue;
                    using (var subTreeReader = xmlReader.ReadSubtree())
                    {
                        var doc = new XmlDocument();
                        doc.Load(subTreeReader);
                        Console.WriteLine("Received XML document: {0}", doc.OuterXml);
                        return;
                    }
                }
            }
        }

    }
}
