﻿using System;
using System.Diagnostics;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Text;
using System.Threading;

namespace NetworkChallengedStreams
{
    class Server
    {
        public void Run()
        {
            try
            {
                Console.WriteLine("[Server] Running server on 127.0.0.1:{0}", Constants.Port);
                var localEndPoint = new IPEndPoint(IPAddress.Loopback, Constants.Port);

                var socket = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
                socket.NoDelay = true;
                socket.Bind(localEndPoint);
                socket.Listen(10);

                while (true)
                {
                    Console.WriteLine("[Server] Waiting for connection...");
                    var handler = socket.Accept();
                    new Thread(HandleConnection).Start(handler);
                }
            }
            catch (Exception ex)
            {
                Console.WriteLine("[Server] Error: " + ex);
            }

        }

        private static void HandleConnection(object socketObj)
        {
            new ConnectionHandler(socketObj as Socket).RunHandler();
        }

        private class ConnectionHandler
        {
            private readonly int _threadId;
            private readonly Socket _socket;

            public ConnectionHandler(Socket socket)
            {
                _threadId = Thread.CurrentThread.ManagedThreadId;
                _socket = socket;
            }

            public void RunHandler()
            {
                Log("Received incoming connection");

                if (_socket == null)
                {
                    Log("Socket object is invalid, exiting connection handler");
                }

                Debug.Assert(_socket != null, "_socket != null");

                try
                {
                    using (_socket)
                    {
                        using (var stream = new NetworkStream(_socket))
                        using (var writer = new StreamWriter(stream, new UTF8Encoding(false), 8192))
                        {
                            int input = stream.ReadByte();
                            if (input == -1)
                            {
                                _socket.Shutdown(SocketShutdown.Send);
                                Log("Connection closed");
                                return;
                            }

                            Log("Running test scenario {0}", input);
                            RunTestScenario(input, writer);

                            stream.ReadByte();
                            _socket.Shutdown(SocketShutdown.Send);
                            Log("Received client Ack. Connection closed");
                        }
                    }
                }
                catch (Exception ex)
                {
                    Log("Error handling connection: {0}", ex);
                }
            }

            private void Log(string format, params object[] args)
            {
                var s = String.Format(format, args);
                Console.WriteLine("[Server] #{0} {1}", _threadId, s);
            }

            private void RunTestScenario(int input, TextWriter writer)
            {
                switch (input)
                {
                    case 1:
                        SendOneMessage(writer, 512);
                        break;

                    case 2:
                        SendChunks(writer, 1023, true);
                        break;

                    case 3:
                        SendChunks(writer, 4095, true);
                        break;

                    case 4:
                        SendChunks(writer, 4048, false);
                        break;

                    default:
                        throw new ApplicationException("Test scenario not supported: " + input);
                }
            }

            private void SendOneMessage(TextWriter writer, int size)
            {
                string s = "<?xml version='1.0'?><message>" + new string('a', size) + "</message>";
                Log("Sending {0} characters", s.Length);
                writer.Write(s);
                writer.Flush();
            }

            private void SendChunks(TextWriter writer, int size, bool sendPound)
            {
                writer.Write("<?xml version='1.0' encoding='utf-8'?><message>");
                writer.Flush();
                Thread.Sleep(100);

                writer.Write(new string('a', size));
                if (sendPound) writer.Write("\u00A3"); // pound sign
                writer.Write("</message>"); 
                writer.Flush();
            }
        }
    }
}
